//-----------------------------------------------------------------------------
// File: Graphics.cpp
//
// Copyright (c) Astralax. All rights reserved.
// Author: Sedov Alexey
//-----------------------------------------------------------------------------

#include "stdafx.h"
#include "Graphics.h"
#include "Game.h"
#include "stb_image.h"
#include <math.h>

float convert_component255_to_float[256];

void SURFACE_RGB::Color32ToRgb(unsigned int color)
{
	unsigned short float_index=color & 0xff;
	b=convert_component255_to_float[float_index];

	float_index=(color>>8) & 0xff;
	g=convert_component255_to_float[float_index];

	float_index=(color>>16) & 0xff;
	r=convert_component255_to_float[float_index];
}

unsigned int SURFACE_RGB::RgbToColor32()
{
	unsigned int color=(unsigned int)(b*255.f);
	unsigned int c=((unsigned int)(g*255.f))<<8;
	color=color | c;
	c=((unsigned int)(r*255.f))<<16;
	color=color | c;
	return color;
}

void SURFACE_RGB::BlendNormal(SURFACE_RGB* rgb, float alpha)
{
	float alpha_dest=1.f-alpha;
	r=r*alpha_dest+rgb->r*alpha;
	g=g*alpha_dest+rgb->g*alpha;
	b=b*alpha_dest+rgb->b*alpha;
}

void SURFACE_RGB::Mul(float alpha)
{
	r*=alpha;
	g*=alpha;
	b*=alpha;
}

void SURFACE_RGB::Interpolation(SURFACE_RGB* rgb, float factor)
{
	float rr=rgb->r-r;
	float rg=rgb->g-g;
	float rb=rgb->b-b;

	r+=rr*factor;
	g+=rg*factor;
	b+=rb*factor;
}

// -----------------------------------------------------------

//   TextBorder

TextBorder::TextBorder()
{
	k_pset=0;
	m_pset=NULL;
	size=0;
}

TextBorder::~TextBorder()
{
	if (m_pset)
		delete []m_pset;
}

void TextBorder::Clear()
{
	k_pset=0;
	if (m_pset)
	{
		delete []m_pset;
		m_pset=NULL;
	}
	size=0;
}

void TextBorder::CreateCircle(int r)
{
	int i;
	Clear();
	GPoint* vm_pset=new GPoint[1000];
	size=r;
	int d=r*2+1;
	for (i=0;i<d;i++)
	{
		for (int j=0;j<d;j++)
		{
			int ir=i-r;
			int jr=j-r;
			double ddist=sqrt((double)(ir*ir+jr*jr));
			int dist=(int)ddist;
			if (ddist-dist>0.5)
				dist++;

			if (dist<=r)
			{
				//   
				GPoint* pt=&(vm_pset[k_pset]);
				pt->x=i-r;
				pt->y=j-r;
				k_pset++;
			}
		}
	}

	m_pset=new GPoint[k_pset];
	for (i=0;i<k_pset;i++)
		m_pset[i]=vm_pset[i];

	delete []vm_pset;
}

// -----------------------------------------------------------

SYMBOL::SYMBOL()
{
	symbol=NULL;
	width=height=0;
}

SYMBOL::~SYMBOL()
{
	if (symbol)
	{
		for (int i=0;i<width;i++)
			delete []symbol[i];

		delete []symbol;

		symbol=NULL;
		width=height=0;
	}
}

//  
//  :
//	alpha      unsigned char,   
//	x,y -   
//	width, height -  
void SYMBOL::CreateSymbol(unsigned char** alpha, int x, int y, int width, int height)
{
	this->width=width;
	this->height=height;

	float factor=1.f/255.f;

	symbol=new float*[width];
	for (int i=0;i<width;i++)
	{
		symbol[i]=new float[height];
		for (int j=0;j<height;j++)
		{
			float a=(float)alpha[i+x][j+y];
			symbol[i][j]=a*factor;
		}
	}
}

// --------------------------------------------------------------------------------------------

GPointList::GPointList()
{
	k_point=0;
	max_point=10;
	m_point=new GPoint[max_point];
}

GPointList::~GPointList()
{
	delete []m_point;
	m_point=NULL;
}

void GPointList::Add(int x, int y)
{
	if (max_point==k_point)
	{
		//   
		int step=max_point/4;
		if (step<4)
			step=4;
		max_point+=step;
		GPoint* vm_point=new GPoint[max_point];

		for (int i=0;i<k_point;i++)
			vm_point[i]=m_point[i];

		delete []m_point;
		m_point=vm_point;
	}

	m_point[k_point].x=x;
	m_point[k_point].y=y;
	k_point++;
}

// --------------------------------------------------------------------------------------------

//   b1 - b2,       a1 - a2
//  false,    
//     1   
bool IntersectInterval(int a1, int a2, int& b1, int& b2)
{
	if (a2<b1 || b2<a1)
		return false;

	if (b1<a1)
		b1=a1;

	if (b2>a2)
		b2=a2;

	return true;
}

GDevice* GSurface::device=NULL;

GSurface::GSurface()
{
	surface=NULL;
	alpha=NULL;
	width=0;
	height=0;

	is_opacity=true;

	blt_alpha_factor=1.f;

	ResetChange();

	clip_x=clip_y=clip_width=clip_height=0;
	offset_x=offset_y=0;

	user=0;
}

GSurface::~GSurface()
{
	Destroy();
}

void GSurface::Destroy()
{
	locked.ptr=NULL;

	if (surface)
	{
		delete []surface;
		surface=NULL;
	}
	if (alpha)
	{
		delete []alpha;
		alpha=NULL;
	}
	width=0;
	height=0;

	ResetChange();
}

//  
void GSurface::GetSize(int& wi, int& he)
{
	wi=width;
	he=height;
}

//   
void GSurface::DrawRect(int x, int y, int width, int height, unsigned int color)
{
	DrawHorizontalLine(x, y, width, color);
	DrawVerticalLine(x+width-1, y, height, color);
	DrawHorizontalLine(x+width-2, y+height-1, -(width-1), color);
	DrawVerticalLine(x, y+height-2, -(height-1), color);
}

//  
bool GSurface::Create(int width, int height)
{
	Destroy();

	is_changed=true;

	int bpp=GetBPP();
	int factor=bpp/8;

	this->width=width;
	this->height=height;

	int len=width*height*factor;

	surface=new unsigned char[len];

	Clear(0);

	ResetClip();

	is_opacity=true;

	return true;
}

//     
bool GSurface::Load(unsigned char* m_data, int k_data)
{
	int comp;

	unsigned char* pixels=stbi_load_from_memory(m_data, k_data, &width, &height, &comp, 4);
	if (pixels)
	{
		unsigned int* pixels4=(unsigned int*)pixels;

		//     pixels  
		Create(width, height);

		int bpp=GetBPP();
		int factor=bpp/8;
		int len=width*height;

		//  RGB
		for (int i=0;i<len;i++)
		{
			unsigned char* p=(unsigned char*)&(pixels4[i]);
			unsigned char v=p[0];
			p[0]=p[2];
			p[2]=v;
			Color32ToColor(pixels4[i], &(surface[i*factor]));
		}

		//   -
		bool is_alpha=false;
		bool is_alpha_full=false;
		unsigned char* pixels_alpha=pixels+3;
		for (int i=0;i<len;i++)
		{
			unsigned char a=pixels_alpha[i*4];
			if (a!=0xff)
			{
				is_alpha=true;
				if (a)
				{
					//  -
					is_alpha_full=true;
					break;
				}
			}
		}

		if (is_alpha_full)
		{
			//    
			is_opacity=false;
			float div255=1.f/255.f;
			alpha=new float[len];
			for (int i=0;i<len;i++)
			{
				unsigned char a=pixels_alpha[i*4];
				alpha[i]=div255*a;
			}
		}
		else
		{
			if (is_alpha)
			{
				//     
				is_opacity=false;
				for (int i=0;i<len;i++)
				{
					unsigned char a=pixels_alpha[i*4];
					if (!a)
					{
						//     
						MaskTransparentPixel(i);
					}
				}
			}
		}

		stbi_image_free(pixels);
		pixels=NULL;

		return true;
	}
	return false;
}

//   
void GSurface::ResetClip()
{
	clip_x=clip_y=0;
	clip_width=width;
	clip_height=height;
}

//   
bool GSurface::SetClip(int x, int y, int width, int height)
{
	ResetClip();
	int bx=x;
	int by=y;
	if (IntersectRect(bx, by, x, y, width, height))
	{
		clip_x=x;
		clip_y=y;
		clip_width=width;
		clip_height=height;
		return true;
	}
	return false;
}

//   
void GSurface::GetClip(int& x, int& y, int& width, int& height)
{
	x=clip_x;
	y=clip_y;
	width=clip_width;
	height=clip_height;
}

//  
SURFACE_LOCKED* GSurface::Lock()
{
	if (!locked.ptr)
	{
		locked.ptr=surface;
		locked.pitch=width*GetBPP()/8;
		locked.alpha_ptr=alpha;
		locked.alpha_pitch=width*4;
	}
	return &locked;
}

//  
void GSurface::Unlock()
{
	if (locked.ptr)
	{
		locked.ptr=NULL;
	}
}

//     
// ,      
void GSurface::SetPtrAndPitch(void* ptr, int pitch)
{
	locked.ptr=ptr;
	locked.pitch=pitch;
	locked.alpha_ptr=NULL;
	locked.alpha_pitch=0;
}

//    ,       
bool GSurface::IntersectRect(int& x, int& y, int& x_from, int& y_from, int& width, int& height)
{
	int b1=x;
	int b2=b1+width-1;
	int b22=b2;
	//if (!IntersectInterval(0, this->width-1, b1, b2))
	if (!IntersectInterval(clip_x, clip_x+clip_width-1, b1, b2))
		return false;
	int r1=b1-x;
	int r2=b22-b2;
	x_from+=r1;
	width-=r1;
	width-=r2;
	x+=r1;

	b1=y;
	b2=b1+height-1;
	b22=b2;
	//if (!IntersectInterval(0, this->height-1, b1, b2))
	if (!IntersectInterval(clip_y, clip_y+clip_height-1, b1, b2))
		return false;
	r1=b1-y;
	r2=b22-b2;
	y_from+=r1;
	height-=r1;
	height-=r2;
	y+=r1;

	return true;
}

//  
//  :
//	- 
//	- 
//	-  
//	- 
bool GSurface::TextOut(int x, int y, const char* text, int font, unsigned int color, unsigned int color_transparent)
{
	if (device)
	{
		GFontex* fontex=device->GetFontex(font);
		if (fontex)
		{
			int i=0;
			while (true)
			{
				unsigned char s=(unsigned char)text[i];
				if (!s)
					break;
				i++;
				if (s>=32)
				{
					SYMBOL* symbol=&(fontex->symbols[s-32]);
					BltSymbol(symbol, x, y, color, color_transparent);
					x+=symbol->width;
				}
			}
			return true;
		}
	}
	return false;
}

//  32-    
void GSurface::Color32ToColor(unsigned int color32, void* color)
{
	SURFACE_RGB rgb;
	rgb.Color32ToRgb(color32);
	RgbToPixel(&rgb, color);
}

//      
void GSurface::ChangeAlpha(float dalpha)
{
	int len=width*height;
	if (!alpha)
	{
		//     -
		float* a=new float[len];
		if (IsOpacity())
		{
			//   
			for (int i=0;i<len;i++)
				a[i]=1.f;
			is_opacity=false;
		}
		else
		{
			// ,     
			for (int i=0;i<width;i++)
			{
				for (int j=0;j<height;j++)
				{
					int index=i+j*width;
					if (IsOver(i, j))
						a[index]=1.f;
					else
						a[index]=0.f;
				}
			}
		}
		alpha=a;
	}

	//     
	if (dalpha>0)
	{
		for (int i=0;i<len;i++)
		{
			float a=alpha[i];
			if (a)
			{
				a+=dalpha;
				if (a>1.f)
					a=1.f;
				alpha[i]=a;
			}
		}
	}
	else
	{
		for (int i=0;i<len;i++)
		{
			float a=alpha[i];
			if (a)
			{
				a+=dalpha;
				if (a<0.f)
					a=0.f;
				alpha[i]=a;
			}
		}
	}

	locked.ptr=NULL;
}

////      
//bool GSurface::Blt(GSurface* from, int x, int y, int x_from, int y_from, int width, int height, bool transparent)
//{
//	//      
//	if ((transparent==false || from->is_opacity) && this->width==width && this->height==height && x==x_from && y==y_from && (x | y)==0)
//	{
//		is_changed=true;
//
//		int factor=GetBPP()/8;
//
//		/*
//		if (alpha)
//		{
//			if (from->alpha)
//			{
//				memcpy(surface, from->surface, width*height*factor);
//				//    
//				memcpy(alpha, from->alpha, width*height*4);
//				return true;
//			}
//		}
//		else
//		{
//			if (!from->alpha)
//			{
//				memcpy(surface, from->surface, width*height*factor);
//				return true;
//			}
//		}
//		*/
//
//		memcpy(surface, from->surface, width*height*factor);
//		return true;
//	}
//	return false;
//}

//      
bool GSurface::Blt(GSurface* from, int x, int y, int x_from, int y_from, int width, int height, bool transparent)
{
	x+=offset_x;
	y+=offset_y;

	int wi,he;
	from->GetSize(wi, he);
	if (!width)
		width=wi;
	if (!height)
		height=he;

	if (from->IsOpacity())
		transparent=false;

	if (IntersectRect(x, y, x_from, y_from, width, height))
	{
		is_changed=true;
		if ((transparent==false || from->is_opacity) && this->width==width && this->width==wi && this->height==height && this->height==he && x==x_from && y==y_from && (x | y)==0)
		{
			int factor=GetBPP()/8;
			memcpy(surface, from->surface, width*height*factor);
		}
		else
		{
			SURFACE_LOCKED* to_lock=Lock();
			SURFACE_LOCKED* from_lock=from->Lock();

			int bpp=GetBPP();
			int factor=bpp/8;

			BLT_LINE_DATA bld;
			bld.from=((unsigned char*)from_lock->ptr)+x_from*factor+y_from*from_lock->pitch;
			bld.to=((unsigned char*)to_lock->ptr)+x*factor+y*to_lock->pitch;
			bld.width=width;
			bld.is_opacity=!transparent;
			bld.is_alpha=(from_lock->alpha_ptr!=NULL);
			bld.alpha_factor=1.f;
			if (!bld.is_opacity)
			{
				float blt_alpha_factor=from->GetBltAlphaFactor();
				if (blt_alpha_factor<=1)
				{
					if (blt_alpha_factor<=0.f)
					{
						bld.alpha_factor=0.f;
						return true;
					}
					else
						bld.alpha_factor=blt_alpha_factor;
				}
			}

			int fwi, fhe;
			from->GetSize(fwi, fhe);
			int alpha_pitch=from_lock->alpha_pitch/4;
			bld.alpha=from_lock->alpha_ptr+x_from+y_from*alpha_pitch;

			if (from==this)
			{
				//       
				if (x>x_from)
				{
					//    ,      ,    
					int sx=(width-1)*factor;
					bld.from=((unsigned char*)bld.from)+sx;
					bld.to=((unsigned char*)bld.to)+sx;
					bld.alpha=bld.alpha+(width-1);
					if (y>y_from)
					{
						//    ,      ,    
						bld.from=((unsigned char*)bld.from)+(height-1)*from_lock->pitch;
						bld.to=((unsigned char*)bld.to)+(height-1)*to_lock->pitch;
						bld.alpha=bld.alpha+(height-1)*alpha_pitch;

						//     
						for (int i=0;i<height;i++)
						{
							BltHorizontalLineRev(&bld);
							bld.from=((unsigned char*)bld.from)-from_lock->pitch;
							bld.to=((unsigned char*)bld.to)-to_lock->pitch;
							bld.alpha-=alpha_pitch;
						}
					}
					else
					{
						//     
						for (int i=0;i<height;i++)
						{
							BltHorizontalLineRev(&bld);
							bld.from=((unsigned char*)bld.from)+from_lock->pitch;
							bld.to=((unsigned char*)bld.to)+to_lock->pitch;
							bld.alpha+=alpha_pitch;
						}
					}
				}
				else
				{
					if (y>y_from)
					{
						//    ,      ,    
						bld.from=((unsigned char*)bld.from)+(height-1)*from_lock->pitch;
						bld.to=((unsigned char*)bld.to)+(height-1)*to_lock->pitch;
						bld.alpha=bld.alpha+(height-1)*alpha_pitch;
						//     
						for (int i=0;i<height;i++)
						{
							BltHorizontalLine(&bld);
							bld.from=((unsigned char*)bld.from)-from_lock->pitch;
							bld.to=((unsigned char*)bld.to)-to_lock->pitch;
							bld.alpha-=alpha_pitch;
						}
					}
					else
					{
						//     
						for (int i=0;i<height;i++)
						{
							BltHorizontalLine(&bld);
							bld.from=((unsigned char*)bld.from)+from_lock->pitch;
							bld.to=((unsigned char*)bld.to)+to_lock->pitch;
							bld.alpha+=alpha_pitch;
						}
					}
				}
			}
			else
			{
				//     
				for (int i=0;i<height;i++)
				{
					BltHorizontalLine(&bld);
					bld.from=((unsigned char*)bld.from)+from_lock->pitch;
					bld.to=((unsigned char*)bld.to)+to_lock->pitch;
					bld.alpha+=alpha_pitch;
				}
			}
		}
	}
	return true;
}

//    
void GSurface::FlipHorizontal()
{
	SURFACE_LOCKED* locked=Lock();
	char* ptr=(char*)locked->ptr;

	char v[8];
	int step=GetBPP()/8;

	int wi2=width/2;

	if (locked->alpha_ptr)
	{
		//  -
		int alpha_pitch=locked->alpha_pitch/4;

		for (int j=0;j<height;j++)
		{
			char* ptr1=ptr+j*locked->pitch;
			char* ptr2=ptr1+(width-1)*step;

			float* aptr1=locked->alpha_ptr+j*alpha_pitch;
			float* aptr2=aptr1+(width-1);

			for (int i=0;i<wi2;i++)
			{
				CopyPixel(ptr1, v);
				CopyPixel(ptr2, ptr1);
				CopyPixel(v, ptr2);
				ptr1+=step;
				ptr2-=step;

				float valpha=*aptr1;
				*aptr1=*aptr2;
				*aptr2=valpha;
				aptr1++;
				aptr2--;
			}
		}
	}
	else
	{
		// - 
		for (int j=0;j<height;j++)
		{
			char* ptr1=ptr+j*locked->pitch;
			char* ptr2=ptr1+(width-1)*step;

			for (int i=0;i<wi2;i++)
			{
				CopyPixel(ptr1, v);
				CopyPixel(ptr2, ptr1);
				CopyPixel(v, ptr2);
				ptr1+=step;
				ptr2-=step;
			}
		}
	}

	Unlock();
}

//    
void GSurface::FlipVertical()
{
	SURFACE_LOCKED* locked=Lock();
	char* ptr=(char*)locked->ptr;

	char v[8];
	int step=GetBPP()/8;

	int he2=height/2;

	if (locked->alpha_ptr)
	{
		//  -
		int alpha_pitch=locked->alpha_pitch/4;

		for (int j=0;j<he2;j++)
		{
			char* ptr1=ptr+j*locked->pitch;
			char* ptr2=ptr+(height-j-1)*locked->pitch;

			float* aptr1=locked->alpha_ptr+j*alpha_pitch;
			float* aptr2=locked->alpha_ptr+(height-j-1)*alpha_pitch;

			for (int i=0;i<width;i++)
			{
				CopyPixel(ptr1, v);
				CopyPixel(ptr2, ptr1);
				CopyPixel(v, ptr2);
				ptr1+=step;
				ptr2+=step;

				float valpha=*aptr1;
				*aptr1=*aptr2;
				*aptr2=valpha;
				aptr1++;
				aptr2++;
			}
		}
	}
	else
	{
		// - 
		for (int j=0;j<he2;j++)
		{
			char* ptr1=ptr+j*locked->pitch;
			char* ptr2=ptr+(height-j-1)*locked->pitch;

			for (int i=0;i<width;i++)
			{
				CopyPixel(ptr1, v);
				CopyPixel(ptr2, ptr1);
				CopyPixel(v, ptr2);
				ptr1+=step;
				ptr2+=step;
			}
		}
	}

	Unlock();
}

//        
void GSurface::CreateFromText(const char* str, int font, unsigned  int color, TextBorder* tb, unsigned int border_color, bool hor_vert)
{
	CMagicString text=str;
	int vert_minus_high=5;
	GSize size;
	if (hor_vert)
	{
		//  
		int len=text.GetLength();
		size.cx=0;
		size.cy=0;

		char ss[2];
		ss[1]=0;

		for (int w=0;w<len;w++)
		{
			char simbol=text[w];

			ss[0]=simbol;
			int swi, she;
			device->GetTextSize(ss, font, swi, she);
			GSize vsize(swi, she);

			vsize.cy-=vert_minus_high;
			size.cy+=vsize.cy;
			if (size.cx<vsize.cx)
				size.cx=vsize.cx;
		}
		size.cy+=vert_minus_high;
	}
	else
	{
		//  
		int swi, she;
		const char* str=text.c_str();
		device->GetTextSize(str, font, swi, she);
		size.cx=swi;
		size.cy=she;
	}
	int wi=size.cx+tb->size*2;
	int he=size.cy+tb->size*2;

	Create(wi, he);
	int len=wi*he;
	//   
	for (int i=0;i<len;i++)
		MaskTransparentPixel(i);
	is_opacity=false;

	if (hor_vert)
	{
		//  
		int len=text.GetLength();
		int sy=0;

		char ss[2];
		ss[1]=0;

		for (int w=0;w<len;w++)
		{
			ss[0]=text[w];
			int swi, she;
			device->GetTextSize(ss, font, swi, she);
			GSize vsize(swi, she);

			vsize.cy-=vert_minus_high;
			int sx=(size.cx-vsize.cx)/2;

			TextOut(tb->size+sx,tb->size+sy, ss, font, color);

			sy+=vsize.cy;
		}
	}
	else
	{
		//  
		const char* str=text.c_str();
		TextOut(tb->size, tb->size, str, font, color);
	}

	//        32 
	SURFACE_RGB color_rgb;
	bool is=false;
	for (int k=0;k<wi;k++)
	{
		for (int l=0;l<he;l++)
		{
			if (IsOver(k, l))
			{
				is=true;
				GetPixelRGB(k, l, &color_rgb);
				k=wi;
				break;
			}
		}
	}

	if (is)
	{
		SURFACE_RGB bclr, rgb;
		bclr.Color32ToRgb(border_color);

		SetPixelRGB(0, 0, &bclr);
		GetPixelRGB(0, 0, &bclr);
		MaskTransparentPixel(0);

		for (int k=0;k<wi;k++)
		{
			for (int l=0;l<he;l++)
			{
				if (IsOver(k, l))
				{
					GetPixelRGB(k, l, &rgb);
					if (rgb.r!=bclr.r || rgb.g!=bclr.g || rgb.b!=bclr.b)
					{
						for (int i=0;i<tb->k_pset;i++)
						{
							GPoint* sm=&(tb->m_pset[i]);
							int vx=k+sm->x;
							int vy=l+sm->y;
							if (vx>=0 && vx<wi && vy>=0 && vy<he)
							{
								if (!IsOver(vx, vy))
									SetPixelRGB(vx, vy, &bclr);
							}
						}
					}
				}
			}
		}
	}
}

struct TargaFileHeader
{
	char idLength;
	char colourMapType;
	char dataTypeCode;
	short colourMapOrigin;
	short colourMapLength;
	char colourMapDepth;
	short xOrigin;
	short yOrigin;
	short width;
	short height;
	char bitsPerPixel;
	char imageDescriptor;
};

//    32- 
unsigned char* GSurface::ConvertationToArray32()
{
	int len=width*height;
	unsigned int* pixel=new unsigned int[len];
	SURFACE_RGB rgb;
	int index=0;
	for (int j=0;j<height;j++)
	{
		for (int i=0;i<width;i++)
		{
			unsigned int color;
			float alpha=GetPixelAlpha(i, j);
			if (alpha)
			{
				GetPixelRGB(i, j, &rgb);
				color=rgb.RgbToColor32();
			}
			else
				color=0;

			unsigned char a=(unsigned char)(alpha*255.f);
			int a2=a;
			a2=a2<<24;
			color=color | a2;
			pixel[index]=color;
			index++;
		}
	}

	return (unsigned char*)pixel;
}

// eng: Saves TGA for testing
// rus:  TGA  
bool GSurface::SaveTGA(const char* file)
{
	unsigned char* pixel=ConvertationToArray32();

	//     
	int he2=height/2;
	int y1=0;
	int y2=height-1;
	unsigned int* pixel4=(unsigned int*)pixel;
	for (int i=0;i<he2;i++)
	{
		unsigned int* p1=pixel4+y1*width;
		unsigned int* p2=pixel4+y2*width;
		for (int j=0;j<width;j++)
		{
			unsigned int v=p1[j];
			p1[j]=p2[j];
			p2[j]=v;
		}
		y1++;
		y2--;
	}

	int components=4;
	TargaFileHeader fileHeader;
	memset(&fileHeader, 0, sizeof(TargaFileHeader));
	fileHeader.dataTypeCode=2;
	fileHeader.bitsPerPixel= components*8;
	fileHeader.width=width;
	fileHeader.height=height;
	fileHeader.imageDescriptor= components==4? 8: 0;

	FILE* f=device->platform->OpenFile(file, "wb");
	if (f)
	{
		fwrite(&fileHeader.idLength, sizeof(fileHeader.idLength), 1, f); 
		fwrite(&fileHeader.colourMapType, sizeof(fileHeader.colourMapType), 1, f);
		fwrite(&fileHeader.dataTypeCode, sizeof(fileHeader.dataTypeCode), 1, f);
		fwrite(&fileHeader.colourMapOrigin, sizeof(fileHeader.colourMapOrigin), 1, f);
		fwrite(&fileHeader.colourMapLength, sizeof(fileHeader.colourMapLength), 1, f);
		fwrite(&fileHeader.colourMapDepth, sizeof(fileHeader.colourMapDepth), 1, f);
		fwrite(&fileHeader.xOrigin, sizeof(fileHeader.xOrigin), 1, f);
		fwrite(&fileHeader.yOrigin, sizeof(fileHeader.yOrigin), 1, f);
		fwrite(&fileHeader.width, sizeof(fileHeader.width), 1, f);
		fwrite(&fileHeader.height, sizeof(fileHeader.height), 1, f);
		fwrite(&fileHeader.bitsPerPixel, sizeof(fileHeader.bitsPerPixel), 1, f);
		fwrite(&fileHeader.imageDescriptor, sizeof(fileHeader.imageDescriptor), 1, f);

		fwrite(pixel, 1, width*height*components, f);

		device->platform->CloseFile(f);

		delete []pixel;
		return true;
	}

	delete []pixel;
	return false;
}



// -------------------------------------------------------------

GImage::GImage(GSurface* surface, int pivot_x, int pivot_y)
{
	this->surface=surface;
	this->pivot_x=pivot_x;
	this->pivot_y=pivot_y;
}

void GImage::GetSize(int& wi, int& he)
{
	if (surface)
		surface->GetSize(wi, he);
	else
		wi=he=0;
}

void GImage::GetPivot(int& px, int& py)
{
	px=pivot_x;
	py=pivot_y;
}

//      
bool GImage::IsOver(int x, int y)
{
	if (surface)
	{
		x+=pivot_x;
		y+=pivot_y;
		return surface->IsOver(x, y);
	}
	return false;
}

//    
void GImage::Blt(int x, int y, GSurface* to, bool transparent)
{
	x-=pivot_x;
	y-=pivot_y;

	int wi, he;
	surface->GetSize(wi, he);
	to->Blt(surface, x, y, 0, 0, wi, he, transparent);
}

// -------------------------------------------------------------

GImageList::GImageList()
{
	width=height=0;
	k_image=0;
	m_image=NULL;
}

GImageList::~GImageList()
{
	for (int i=0;i<k_image;i++)
	{
		GSurface::device->DeleteSurface(m_image[i]->GetSurface());
		m_image[i]->SetSurface(NULL);
		delete m_image[i];
	}

	if (m_image)
	{
		delete []m_image;
		m_image=NULL;
	}
	k_image=0;
}

bool GImageList::Create(int width, int height)
{
	if ((!this->width) && width>0 && height>0)
	{
		this->width=width;
		this->height=height;
		return true;
	}
	return false;
}

int GImageList::AddImage(GImage* image)
{
	if (k_image)
	{
		GImage** vm_image=new GImage*[k_image+1];

		for (int i=0;i<k_image;i++)
			vm_image[i]=m_image[i];

		delete []m_image;
		m_image=vm_image;
	}
	else
	{
		m_image=new GImage*[1];
	}

	GSurface* surface=GSurface::device->AddSurface();
	surface->Create(width, height);

	int wi, he;
	image->GetSize(wi, he);
	int x=(width-wi)/2;
	int y=(height-he)/2;
	surface->Copy(image->GetSurface(), x, y);

	int px, py;
	image->GetPivot(px, py);
	m_image[k_image]=new GImage(surface, px, py);
	k_image++;
	return k_image-1;
}

int GImageList::AddImage(GSurface* surface)
{
	GImage* image=GSurface::device->AddPicture(surface, 0, 0);
	return AddImage(image);
}

GImage* GImageList::GetImage(int index)
{
	if (index>=0 && index<k_image)
		return m_image[index];
	return NULL;
}

// -------------------------------------------------------------

GGame* GDevice::game=NULL;
GPlatform* GDevice::platform=NULL;

GDevice::GDevice(bool fullscreen_mode)
{
	this->fullscreen_mode=fullscreen_mode;

	width=0;
	height=0;

	monitor_width=0;
	monitor_height=0;
	changed_monitor_resolution=true;

	k_mode=0;

	started_mode.width=0;
	started_mode.height=0;
	started_mode.bpp=0;

	active_width=0;
	active_height=0;
	active_monitor_width=0;
	active_monitor_height=0;

	k_fontex=0;
	m_fontex=NULL;

	k_surface=0;
	m_surface=NULL;

	k_cursor=0;
	t_cursor=-1;
	m_cursor=NULL;

	k_picture=0;
	m_picture=NULL;

	for (int i=0;i<256;i++)
		convert_component255_to_float[i]=((float)i)/255.f;

	windowed=false;

	tex_rect.left=0;
	tex_rect.top=0;
	tex_rect.right=0;
	tex_rect.bottom=0;
}

GDevice::~GDevice()
{
	int i;
	for (i=0;i<k_fontex;i++)
		delete m_fontex[i];
	if (m_fontex)
	{
		delete []m_fontex;
		m_fontex=NULL;
	}
	k_fontex=0;

	for (i=0;i<k_surface;i++)
	{
		if (DeleteSurface(m_surface[i]))
			i--;
	}

	if (m_surface)
	{
		delete []m_surface;
		m_surface=NULL;
	}
	k_surface=0;

	for (i=0;i<k_cursor;i++)
		delete m_cursor[i];
	if (m_cursor)
	{
		delete []m_cursor;
		m_cursor=NULL;
	}
	k_cursor=0;
	t_cursor=-1;

	for (i=0;i<k_picture;i++)
		delete m_picture[i];
	if (m_picture)
	{
		delete []m_picture;
		m_picture=NULL;
	}
	k_picture=0;
}

//  
bool GDevice::Create(int view_width, int view_height)
{
	Destroy();

	GSurface::device=this;

	VIDEO_MODE cvm;
	GetMonitorMode(&cvm);
	if (!started_mode.width)
	{
		started_mode=cvm;
	}

	if (view_width<=0)
	{
		view_width=cvm.width;
	}
	if (view_height<=0)
	{
		view_height=cvm.height;
	}

	width=view_width;//cvm.width;
	height=view_height;//cvm.height;
	monitor_width=width;
	monitor_height=height;

	ChangeResolution(true, true);
	
	return true;
}

//  
void GDevice::Destroy()
{
	GSurface::device=NULL;
	width=0;
	height=0;
}

// 
void GDevice::GetSize(int& wi, int& he)
{
	wi=width;
	he=height;
}

//    
bool GDevice::IsVideoMode(VIDEO_MODE* vm)
{
	int min_width, min_height;
	GetMinSize(min_width, min_height);
	int max_width, max_height;
	GetMaxSize(max_width, max_height);
	return (vm->bpp==32 && vm->width>=min_width && vm->width<=max_width && vm->height>=min_height&& vm->height<=max_height);
}

//    
void GDevice::RefreshVideoMode()
{
	if (k_mode>0)
	{
		//       
		for (int i=0;i<k_mode;i++)
		{
			int min_index=i;
			VIDEO_MODE* vm=&(m_mode[i]);
			int min_wi=vm->width;
			int min_he=vm->height;
			int min_bpp=vm->bpp;
			for (int j=i+1;j<k_mode;j++)
			{
				vm=&(m_mode[j]);
				bool is=false;
				if (vm->bpp<min_bpp)
					is=true;
				else
				{
					if (vm->bpp==min_bpp)
					{
						if (vm->width<min_wi)
							is=true;
						else
						{
							if (vm->width<min_wi)
								is=true;
							else
							{
								if (vm->width==min_wi)
								{
									if (vm->height<min_he)
										is=true;
								}
							}
						}
					}
				}

				if (is)
				{
					min_index=j;
					min_wi=vm->width;
					min_he=vm->height;
					min_bpp=vm->bpp;
				}
			}

			if (min_index!=i)
			{
				//    
				VIDEO_MODE v;
				v=m_mode[i];
				m_mode[i]=m_mode[min_index];
				m_mode[min_index]=v;
			}
		}
	}
}

//  
bool GDevice::SetVideoMode(VIDEO_MODE* vm)
{
	if (!k_mode)
	{
		//    
		RefreshVideoMode();
	}

	int index=-1;
	if (IsChangedMonitorResolution())
	{
		for (int i=0;i<k_mode;i++)
		{
			if (m_mode[i]==*vm)
			{
				index=i;
				break;
			}
		}
	}
	else
		index=0;

	if (index!=-1)
	{
		bool changed_scene=false;
		if (width!=vm->width || height!=vm->height)
		{
			changed_scene=true;
			width=vm->width;
			height=vm->height;
		}

		bool changed_monitor=false;
		if (IsChangedMonitorResolution())
		{
			if (monitor_width!=vm->width || monitor_height!=vm->height)
			{
				monitor_width=width;
				monitor_height=height;
				changed_monitor=true;
			}
		}

		if (changed_scene || changed_monitor)
		{
			ChangeResolution(changed_monitor, changed_scene);
			if (changed_scene && game)
				game->OnChangeVideoMode(width, height);
		}

		return true;
	}
	return false;
}

//     
int GDevice::GetMaxWidthVideoMode()
{
	RefreshVideoMode();
	int min=0;
	for (int i=0;i<k_mode;i++)
	{
		if (min<m_mode[i].width)
			min=m_mode[i].width;
	}
	return min;
}

//     
int GDevice::GetMaxHeightVideoMode()
{
	RefreshVideoMode();
	int min=0;
	for (int i=0;i<k_mode;i++)
	{
		if (min<m_mode[i].height)
			min=m_mode[i].height;
	}
	return min;
}

CMagicString GetFontParam(const CMagicString& description, int pos)
{
	CMagicString val;
	while (true)
	{
		char s=description[pos];
		if (s=='\r' || s==0)
			return val;
		val+=s;
		pos++;
	}
}

CMagicString FindFontParam(const CMagicString& description, const CMagicString& param)
{
	CMagicString val;
	CMagicString param2=param;
	param2+='=';
	int pos=description.Find(param2);
	if (pos!=-1)
	{
		pos+=param2.GetLength();
		val=GetFontParam(description, pos);
	}
	return val;
}

//  
//  
	//	alpha_texture -       -
	//	description -     
int GDevice::AddFontex(unsigned char** alpha_texture, const CMagicString& description)
{
	//    
	FONTEX_DATA fd;

	CMagicString param="Font";
	CMagicString val=FindFontParam(description, param);
	if (val.IsEmpty())
		return -1;
	fd.name=val;

	param="Height";
	val=FindFontParam(description, param);
	if (val.IsEmpty())
		return -1;
	fd.height=atoi(val);

	param="Weight";
	val=FindFontParam(description, param);
	if (val.IsEmpty())
		return -1;
	fd.weight=atoi(val);

	param="Italic";
	val=FindFontParam(description, param);
	if (val.IsEmpty())
		return -1;
	if (atoi(val))
		fd.italic=true;
	else
		fd.italic=false;

	param="x,y,width";
	int pos=description.Find(param);
	if (pos==-1)
		return false;
	
	//       
	for (int i=0;i<k_fontex;i++)
	{
		if (fd==m_fontex[i]->data)
			return i;
	}

	if (k_fontex)
	{
		GFontex** vm_fontex=new GFontex*[k_fontex+1];

		for (int i=0;i<k_fontex;i++)
			vm_fontex[i]=m_fontex[i];

		delete []m_fontex;
		m_fontex=vm_fontex;
	}
	else
	{
		m_fontex=new GFontex*[1];
	}

	k_fontex++;
	GFontex* new_fontex=new GFontex();
	m_fontex[k_fontex-1]=new_fontex;

	new_fontex->data=fd;

	//   
	for (int i=32;i<256;i++)
	{
		pos=description.Find('=', pos);
		if (pos==-1)
			return false;
		pos++;
		
		if (description[pos]=='=')
		{
			//      
			//    
			pos++;
		}

		param=GetFontParam(description, pos);
		if (param.IsEmpty())
			return false;

		int pos2=param.Find(',');
		if (pos2==-1)
			return false;
		CMagicString p=param.Left(pos2);
		int x=atoi(p);

		pos2++;
		int pos3=param.Find(',', pos2);
		if (pos3==-1)
			return false;
		p=param.Mid(pos2, pos3-pos2);
		int y=atoi(p);

		p=GetFontParam(param, pos3+1);
		if (p.IsEmpty())
			return false;
		int width=atoi(p);

		//   
		SYMBOL* symbol=&(new_fontex->symbols[i-32]);
		symbol->CreateSymbol(alpha_texture, x, y, width, fd.height);
	}

	return k_fontex-1;
}

int GDevice::AddFontex(unsigned char* texture_data, int texture_length,  unsigned char* description_data, int description_length)
{
	int index=-1;

	int comp;
	int wi,he;
	unsigned char* pixels=stbi_load_from_memory(texture_data, texture_length, &wi, &he, &comp, 4);
	if (pixels)
	{
		unsigned char* pixels4=pixels+3;
		unsigned char** alpha=new unsigned char*[wi];
		for (int i=0;i<wi;i++)
			alpha[i]=new unsigned char[he];

		for (int j=0;j<he;j++)
		{
			for (int i=0;i<wi;i++)
			{
				alpha[i][j]=*pixels4;
				pixels4+=4;
			}
		}

		stbi_image_free(pixels);

		CMagicString description(description_length);
		char* s=(char*)description.c_str();
		for (int i=0;i<description_length;i++)
			s[i]=description_data[i];

		index=AddFontex(alpha, description);
		
		for (int i=0;i<wi;i++)
			delete [](alpha[i]);
		delete []alpha;
	}
	return index;
}

//    
GFontex* GDevice::GetFontex(int index)
{
	if (index>=0 && index<k_fontex)
		return m_fontex[index];
	return NULL;
}

//   
void GDevice::GetTextSize(const char* text, int font, int& wi, int& he)
{
	wi=he=0;
	if (text)
	{
		GFontex* fontex=GetFontex(font);
		if (fontex)
		{
			he=fontex->data.height;
			int i=0;
			while (true)
			{
				unsigned char s=(unsigned char)text[i];
				if (!s)
					break;
				i++;
				SYMBOL* symbol=&(fontex->symbols[s-32]);
				wi+=symbol->width;
			}
		}
	}
}

//  
GSurface* GDevice::AddSurface()
{
	if (k_surface)
	{
		GSurface** vm_surface=new GSurface*[k_surface+1];

		for (int i=0;i<k_surface;i++)
			vm_surface[i]=m_surface[i];

		delete []m_surface;
		m_surface=vm_surface;
	}
	else
	{
		m_surface=new GSurface*[1];
	}

	GSurface* surface=NewSurface();
	m_surface[k_surface]=surface;
	k_surface++;

	return surface;
}

//  
bool GDevice::DeleteSurface(GSurface* surface)
{
	int index=-1;
	for (int i=0;i<k_surface;i++)
	{
		if (m_surface[i]==surface)
		{
			index=i;
			break;
		}
	}
	
	if (index!=-1)
	{
		surface->Destroy();
		delete surface;
		if (k_surface>1)
		{
			GSurface** vm_surface=new GSurface*[k_surface-1];
			for (int i=0;i<index;i++)
				vm_surface[i]=m_surface[i];

			for (int i=index+1;i<k_surface;i++)
				vm_surface[i-1]=m_surface[i];

			delete []m_surface;
			m_surface=vm_surface;
		}
		else
		{
			delete []m_surface;
			m_surface=NULL;
		}

		k_surface--;

		return true;
	}
	return false;

}

//  
int GDevice::AddCursor(GSurface* cursor, int pivot_x, int pivot_y)
{
	if (k_cursor)
	{
		GImage** vm_cursor=new GImage*[k_cursor+1];

		for (int i=0;i<k_cursor;i++)
			vm_cursor[i]=m_cursor[i];

		delete []m_cursor;
		m_cursor=vm_cursor;
	}
	else
	{
		m_cursor=new GImage*[1];
	}

	m_cursor[k_cursor]=new GImage(cursor, pivot_x, pivot_y);
	if (!k_cursor)
		t_cursor=0;
	k_cursor++;
	return k_cursor-1;
}

// 
GImage* GDevice::GetCursor()
{
	if (t_cursor!=-1)
		return m_cursor[t_cursor];
	return NULL;
}

bool GDevice::SetCursor(int index)
{
	if (index>=0 && index<k_cursor)
	{
		t_cursor=index;
		return true;
	}
	return false;
}

//  
GImage* GDevice::AddPicture(GSurface* surface, int pivot_x, int pivot_y)
{
	if (k_picture)
	{
		GImage** vm_picture=new GImage*[k_picture+1];

		for (int i=0;i<k_picture;i++)
			vm_picture[i]=m_picture[i];

		delete []m_picture;
		m_picture=vm_picture;
	}
	else
	{
		m_picture=new GImage*[1];
	}

	GImage* picture=new GImage(surface, pivot_x, pivot_y);
	m_picture[k_picture]=picture;
	k_picture++;
	return picture;
}

//   
void GDevice::GetSceneRect(CMagicRect* rect)
{
	rect->left=0;
	rect->top=0;
	rect->right=width;
	rect->bottom=height;
}

//  
void GDevice::Activate()
{
	if (active_monitor_width)
	{
		width=active_width;
		height=active_height;
		monitor_width=active_monitor_width;
		monitor_height=active_monitor_height;
		ChangeResolution(true, true);
		if (game)
			game->OnChangeVideoMode(width, height);
	}
}

//  
void GDevice::Inactivate()
{
	active_width=width;
	active_height=height;
	active_monitor_width=monitor_width;
	active_monitor_height=monitor_height;

	//SetVideoMode(&started_mode);

	monitor_width=started_mode.width;
	monitor_height=started_mode.height;
	SetMonitorResolutionOnly();
}

//        
//    ,       
bool GDevice::CalcTexRect()
{
	bool filter=false;
	int swi=width;
	int she=height;
	double k=((double)swi)/((double)she);
	if (swi>monitor_width)
	{
		swi=monitor_width;
		she=(int)(swi/k);
		filter=true;
	}
	if (she>monitor_height)
	{
		she=monitor_height;
		swi=(int)(she*k);
		filter=true;
	}

	tex_rect.left=(monitor_width-swi)/2;
	tex_rect.top=(monitor_height-she)/2;
	tex_rect.right=tex_rect.left+swi-1;
	tex_rect.bottom=tex_rect.top+she-1;

	return filter;
}