//-----------------------------------------------------------------------------
// File: Graphics.h
//
// Copyright (c) Astralax. All rights reserved.
// Author: Sedov Alexey
//-----------------------------------------------------------------------------

#ifndef ONIMOD_GRAPHICS
#define ONIMOD_GRAPHICS

//#define MAX_FPS

#include "utils.h"

#define COLOR_KEY_EMPTY 0x230955

class GGame;
class GPlatform;
class GDevice;

struct SURFACE_LOCKED
{
	void* ptr;		//       (NULL,    )
	int pitch;		//    

	float* alpha_ptr;	//   
	int alpha_pitch;	//  -  

	SURFACE_LOCKED()
	{
		ptr=NULL;
		pitch=0;

		alpha_ptr=NULL;
		alpha_pitch=0;
	}
};

struct SURFACE_RGB
{
	float r;
	float g;
	float b;

	void Color32ToRgb(unsigned int color);
	unsigned int RgbToColor32();

	void BlendNormal(SURFACE_RGB* rgb, float alpha);

	void Mul(float alpha);

	void Interpolation(SURFACE_RGB* rgb, float factor);
};

// ,    
struct SYMBOL
{
	int width, height;
	float** symbol;

	SYMBOL();
	virtual ~SYMBOL();

	//  
	//  :
	//	alpha -      unsigned char,   
	//	x,y -   
	//	width, height -  
	void CreateSymbol(unsigned char** alpha, int x, int y, int width, int height);
};

struct FONTEX_DATA
{
	CMagicString name;
	int height;
	int weight;
	bool italic;

	bool operator==(const FONTEX_DATA& fd)
	{
		return (name==fd.name && height==fd.height && weight==fd.weight && italic==fd.italic);
	}
};

// ,  
struct GFontex
{
	FONTEX_DATA data;
	SYMBOL symbols[256-32];
};

struct GSize
{
	int cx;
	int cy;

	GSize(){}
	GSize(int wi, int he){cx=wi, cy=he;}
};

struct GPoint
{
	int x, y;
};

class GPointList
{
protected:
	int k_point;
	int max_point;
	GPoint* m_point;

public:
	GPointList();
	virtual ~GPointList();

	void Add(int x, int y);

	GPoint Get(int index){return m_point[index];}
	void Set(int index, GPoint point){m_point[index]=point;}
	int GetCount(){return k_point;}

	void Clear(){k_point=0;}
};

//       
struct TextBorder
{
	int k_pset;
	GPoint* m_pset;
	int size;

	TextBorder();
	~TextBorder();

	void Clear();

	void CreateCircle(int r);
};

struct BLT_LINE_DATA
{
	void* from;
	void* to;
	int width;
	float* alpha;
	bool is_opacity;
	bool is_alpha;
	float alpha_factor;
};

//     ,    
class GSurface
{
protected:
	unsigned char* surface;	//  RGB-
	float* alpha;			//  - (    ,   )
	int width, height;		//  
	SURFACE_LOCKED locked;	//    

	bool is_opacity;		//  ,       

	bool is_changed;		//  ,     (   - )

	float blt_alpha_factor;		// ,        

	int clip_x, clip_y, clip_width, clip_height;	//  
	int offset_x, offset_y;	//    (   )

	int user;

public:
	static GDevice* device;

public:
	GSurface();
	virtual ~GSurface();

	//  
	void GetSize(int& wi, int& he);

	//    
	virtual void FillRect(int x, int y, int width, int height, unsigned int color)=0;

	//   
	virtual void DrawHorizontalLine(int x, int y, int length, unsigned int color)=0;
	//   
	virtual void DrawVerticalLine(int x, int y, int length, unsigned int color)=0;
	//  
	virtual void DrawLine(int x1, int y1, int x2, int y2, unsigned int color)=0;

	//   
	virtual void DrawRect(int x, int y, int width, int height, unsigned int color);

	//   
	virtual void FloodFill(int x, int y, unsigned int fill_color, unsigned int obstacle_color)=0;

	//  ,      
	bool IsOpacity(){return is_opacity;}

	//      
	virtual bool IsOver(int x, int y)=0;

	//  
	bool Create(int width, int height);
	//     
	bool Load(unsigned char* m_data, int k_data);

	//      ,    -
	virtual void MaskTransparentPixel(int position)=0;

	//   RGB 
	virtual void GetPixelRGB(int x, int y, SURFACE_RGB* rgb)=0;

	//   RGB 
	virtual void SetPixelRGB(int x, int y, SURFACE_RGB* rgb)=0;

	//   Alpha 
	virtual float GetPixelAlpha(int x, int y)=0;

	//    
	virtual void Clear(unsigned int color)=0;

	//  
	virtual void Destroy();

	//   
	bool IsChanged(){return is_changed;}
	void ResetChange(bool update=false){is_changed=update;}

	//   
	void ResetClip();

	//   
	bool SetClip(int x, int y, int width, int height);

	//   
	void GetClip(int& x, int& y, int& width, int& height);

	//   
	void SetOffset(int x, int y){offset_x=x; offset_y=y;}

	//   
	void GetOffset(int& x, int& y){x=offset_x; y=offset_y;}

	//  
	SURFACE_LOCKED* Lock();
	//  
	void Unlock();
	//     
	// ,      
	void SetPtrAndPitch(void* ptr, int pitch);

	//    ,       
	bool IntersectRect(int& x, int& y, int& x_from, int& y_from, int& width, int& height);

	//  
	//  :
	//	- 
	//	- 
	//	-  
	//	- 
	bool TextOut(int x, int y, const char* text, int font, unsigned int color, unsigned int color_transparent=COLOR_KEY_EMPTY);

	//  32-    
	void Color32ToColor(unsigned int color32, void* color);

	//   
	virtual int GetBPP()=0;

	//      
	virtual bool Blt(GSurface* from, int x=0, int y=0, int x_from=0, int y_from=0, int width=0, int height=0, bool transparent=true);

	//     
	virtual void BltSymbol(SYMBOL* symbol, int x, int y, unsigned int color, unsigned int color_transparent=COLOR_KEY_EMPTY)=0;

	//    -    (     )
	virtual bool Copy(GSurface* from, int x=0, int y=0)=0;

	//    
	void FlipHorizontal();
	//    
	void FlipVertical();

	//  
	virtual void CopyPixel(void* from, void* to)=0;

	//   32-        (-  )
	//     to_x, to_y.     width, height    
	virtual void BltVideo(unsigned int* from, int to_x, int to_y, int from_x, int from_y, int width, int height, int pitch=0)=0;

	//      
	void ChangeAlpha(float dalpha);

	// ,        
	void SetBltAlphaFactor(float alpha){blt_alpha_factor=alpha;}
	float GetBltAlphaFactor(){return blt_alpha_factor;}

	//       R,G,B
	virtual void PixelToRgb(void* color_ptr, SURFACE_RGB* rgb)=0;
	//    R,G,B  
	virtual void RgbToPixel(SURFACE_RGB* rgb, void* color_ptr)=0;

	//      
	// ,      
	virtual void CopyToTexture32(void* texture_ptr, int texture_pitch)=0;

	//         
	// ,      
	virtual void CopyToTexture(void* texture_ptr, int texture_pitch)=0;

	//        
	void CreateFromText(const char* str, int font, unsigned  int color, TextBorder* tb, unsigned int border_color, bool hor_vert);

	//    32- 
	unsigned char* ConvertationToArray32();

	// eng: Saves TGA for testing
	// rus:  TGA  
	bool SaveTGA(const char* file);

	//  
	void SetUserData(int data){user=data;}
	int GetUserData(){return user;}

protected:
	//      
	virtual void BltHorizontalLine(BLT_LINE_DATA* data)=0;
	//      
	virtual void BltHorizontalLineRev(BLT_LINE_DATA* data)=0;
};

struct VIDEO_MODE
{
	int width;
	int height;
	int bpp;

	bool operator==(const VIDEO_MODE& vm)
	{
		return (width==vm.width && height==vm.height && bpp==vm.bpp);
	}
};

class GImage
{
protected:
	GSurface* surface;
	int pivot_x, pivot_y;

public:
	GImage(GSurface* surface, int pivot_x, int pivot_y);

	void GetSize(int& wi, int& he);

	void GetPivot(int& px, int& py);

	GSurface* GetSurface(){return surface;}
	void SetSurface(GSurface* surface){this->surface=surface;}

	//      
	bool IsOver(int x, int y);

	//    
	void Blt(int x, int y, GSurface* to, bool transparent=false);
};

// ,      
class GImageList
{
protected:
	int width, height;
	int k_image;
	GImage** m_image;

public:
	GImageList();
	virtual ~GImageList();

	bool Create(int width, int height);

	void GetSize(int& wi, int& he){wi=width; he=height;}

	int AddImage(GImage* image);
	int AddImage(GSurface* surface);

	GImage* GetImage(int index);

	int GetImageCount(){return k_image;}
};

// ,      
class GDevice
{
protected:
	int width, height;
	int monitor_width, monitor_height;
	bool changed_monitor_resolution;

	int k_mode;
	VIDEO_MODE m_mode[200];

	VIDEO_MODE started_mode;			// ,      

	int active_width, active_height;						//  ,     
	int active_monitor_width, active_monitor_height;		//  ,     

	int k_fontex;
	GFontex** m_fontex;

	int k_surface;
	GSurface** m_surface;

	int k_cursor;
	int t_cursor;
	GImage** m_cursor;

	int k_picture;
	GImage** m_picture;

	bool windowed;	//   
	bool fullscreen_mode;	//   

	CMagicRect tex_rect; // ,       . right  bottom   ,   

public:
	static GGame* game;
	static GPlatform* platform;

public:
	GDevice(bool fullscreen_mode);
	virtual ~GDevice();

	//  
	virtual bool Create(int view_width, int view_height);

	//  
	virtual void Destroy();

	// 
	void GetSize(int& wi, int& he);

	//  
	int GetVideoModeCount(){return k_mode;}
	VIDEO_MODE* GetVideoMode(int index){return &(m_mode[index]);}

	//  
	virtual bool SetVideoMode(VIDEO_MODE* vm);

	//     
	int GetMaxWidthVideoMode();
	//     
	int GetMaxHeightVideoMode();

protected:
	//    
	virtual bool IsVideoMode(VIDEO_MODE* vm);

public:
	//    
	virtual void RefreshVideoMode();

	//    
	virtual void GetMonitorMode(VIDEO_MODE* vm)=0;

	//    
	virtual void GetScreenModeDefault(VIDEO_MODE* vm)=0;

protected:
	//    /
	virtual void ChangeResolution(bool monitor, bool scene)=0;

	//      (    )
	//    monitor_width, monitor_height
	//    ,   
	virtual bool SetMonitorResolutionOnly()=0;

public:
	void GetStartedMode(VIDEO_MODE* vm){*vm=started_mode;}

	//      
	virtual void GetMinSize(int& min_width, int& min_height){min_width=0; min_height=0;}
	virtual void GetMaxSize(int& max_width, int& max_height){max_width=100000; max_height=100000;}

	//  
	//  
	//	alpha_texture -       -
	//	description -     
	int AddFontex(unsigned char** alpha_texture, const CMagicString& description);
	int AddFontex(unsigned char* texture_data, int texture_length,  unsigned char* description_data, int description_length);

	//    
	GFontex* GetFontex(int index);

	//   
	void GetTextSize(const char* text, int font, int& wi, int& he);

	//  
	GSurface* AddSurface();

	//  
	bool DeleteSurface(GSurface* surface);

	//   
	int GetSurfaceCount(){return k_surface;}

	//  
	int AddCursor(GSurface* cursor, int pivot_x, int pivot_y);

	// 
	GImage* GetCursor();
	bool SetCursor(int index);

	//  
	GImage* AddPicture(GSurface* surface, int pivot_x=0, int pivot_y=0);

	//   
	virtual void GetSceneRect(CMagicRect* rect);

	// 
	virtual void Render()=0;

	//  
	virtual GSurface* NewSurface()=0;

	virtual void LockBackbuffer(SURFACE_LOCKED* locked)=0;
	virtual void UnlockBackbuffer()=0;

	//   
	virtual bool IsLost()=0;

	//  
	virtual void Activate();

	//  
	virtual void Inactivate();

	//    (         )
	virtual bool IsVisibleCursor()=0;

	//    
	bool IsChangedMonitorResolution(){return changed_monitor_resolution;}
	void ChangeMonitorResolution(bool change){changed_monitor_resolution=change;}

	//   
	bool IsWindowed(){return windowed;}

	//        
	//    ,       
	bool CalcTexRect();
	//   
	CMagicRect* GetTexRect(){return &tex_rect;}
};

#endif